package tools.descartes.prisma.agent.wildfly.interceptors;

import java.util.Deque;

import javax.servlet.http.HttpServletRequest;
import javax.xml.namespace.QName;
import javax.xml.ws.handler.MessageContext;

import org.jboss.ws.api.handler.GenericSOAPHandler;

import io.undertow.servlet.spec.HttpServletRequestImpl;
import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.agent.wildfly.callstack.ThreadCallStack;
import tools.descartes.prisma.agent.wildfly.service.MonitoringService;

public class WebServiceIncomingInterceptor extends GenericSOAPHandler {
	
	@Override
	protected boolean handleInbound(MessageContext msgContext) {
		String path = (String)msgContext.get(MessageContext.PATH_INFO);
		QName service = (QName)msgContext.get(MessageContext.WSDL_SERVICE);
		QName interf = (QName)msgContext.get(MessageContext.WSDL_INTERFACE);
		QName operation = (QName)msgContext.get(MessageContext.WSDL_OPERATION);

		Deque<InvocationGroupRecord> threadCallStack = ThreadCallStack.get();
		ComponentRecord crec = MonitoringService.INSTANCE.getTraceRoot().requireComponent(path);
		String componentClass = service.getNamespaceURI();
		if (service.getLocalPart() != null) {
			componentClass = componentClass + service.getLocalPart();
		}
		crec.setComponentClass(service.getNamespaceURI() + service.getLocalPart());
		InterfaceRecord irec = crec.requireInterface(interf.getNamespaceURI() + "/" + interf.getLocalPart());
		OperationRecord orec = irec.requireOperation(operation.getLocalPart());
		InvocationGroupRecord grec = orec.requireInvocationGroup(0);
		
		threadCallStack.push(grec);
		return true;
	}
	
	@Override
	protected boolean handleOutbound(MessageContext msgContext) {
		long endTime = System.nanoTime();
		Deque<InvocationGroupRecord> threadCallStack = ThreadCallStack.get();
		if (!threadCallStack.isEmpty()) {
			InvocationGroupRecord grec = threadCallStack.pop();
			HttpServletRequestImpl request = (HttpServletRequestImpl)msgContext.get(MessageContext.SERVLET_REQUEST);
			long duration = endTime - request.getExchange().getRequestStartTime();
			grec.addInvocation(duration, 0, 0);
		}
		return true;
	}
}
