package tools.descartes.prisma.agent.wildfly.service;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

import org.jboss.logging.Logger;
import org.jboss.msc.service.Service;
import org.jboss.msc.service.ServiceName;
import org.jboss.msc.service.StartContext;
import org.jboss.msc.service.StartException;
import org.jboss.msc.service.StopContext;
import org.jboss.msc.value.InjectedValue;

import edu.kit.ipd.descartes.mm.applicationlevel.functions.DoubleLiteral;
import edu.kit.ipd.descartes.mm.applicationlevel.functions.FunctionsFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.functions.IntSample;
import edu.kit.ipd.descartes.mm.applicationlevel.functions.IntSampleList;
import edu.kit.ipd.descartes.mm.applicationlevel.functions.ProbabilityMassFunction;
import edu.kit.ipd.descartes.mm.applicationlevel.functions.RandomVariable;
import edu.kit.ipd.descartes.mm.applicationlevel.parameterdependencies.ModelVariableCharacterizationType;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.BasicComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Interface;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceProvidingRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Repository;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCall;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.FineGrainedBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ForkAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.InternalAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopIterationCount;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ResourceDemand;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import edu.kit.ipd.descartes.mm.deployment.Deployment;
import edu.kit.ipd.descartes.mm.deployment.DeploymentContext;
import edu.kit.ipd.descartes.mm.deployment.DeploymentFactory;
import edu.kit.ipd.descartes.mm.resourceconfiguration.SchedulingPolicy;
import edu.kit.ipd.descartes.mm.resourcelandscape.ComputingInfrastructure;
import edu.kit.ipd.descartes.mm.resourcelandscape.DataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.DistributedDataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.HardwareInfrastructure;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceType;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceTypeRepository;
import tools.descartes.prisma.agent.wildfly.callstack.ComponentRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InterfaceRecord;
import tools.descartes.prisma.agent.wildfly.callstack.InvocationGroupRecord;
import tools.descartes.prisma.agent.wildfly.callstack.OperationRecord;
import tools.descartes.prisma.core.ApplicationModel;
import tools.descartes.prisma.core.RepositoryConnection;
import tools.descartes.prisma.core.templates.ModelTemplate;
import tools.descartes.prisma.core.templates.deployment.DeploymentTemplate;
import tools.descartes.prisma.core.templates.repository.AssemblyContextTemplate;
import tools.descartes.prisma.core.templates.repository.BasicComponentTemplate;
import tools.descartes.prisma.core.templates.repository.CompositeComponentTemplate;
import tools.descartes.prisma.core.templates.repository.InterfaceProvidingRoleTemplate;
import tools.descartes.prisma.core.templates.repository.InterfaceRequiringRoleTemplate;
import tools.descartes.prisma.core.templates.repository.InterfaceTemplate;
import tools.descartes.prisma.core.templates.repository.ProvidingDelegationConnectorTemplate;
import tools.descartes.prisma.core.templates.repository.RepositoryComponentTemplate;
import tools.descartes.prisma.core.templates.repository.RepositoryTemplate;
import tools.descartes.prisma.core.templates.repository.RequiringDelegationConnectorTemplate;
import tools.descartes.prisma.core.templates.repository.SignatureTemplate;
import tools.descartes.prisma.core.templates.resourceconfiguration.ActiveResourceSpecificationTemplate;
import tools.descartes.prisma.core.templates.resourcelandscape.ComputingInfrastructureTemplate;
import tools.descartes.prisma.core.templates.resourcelandscape.DataCenterTemplate;
import tools.descartes.prisma.core.templates.resourcelandscape.DistributedDataCenterTemplate;
import tools.descartes.prisma.core.templates.resourcetype.ProcessingResourceTypeTemplate;
import tools.descartes.prisma.core.templates.resourcetype.ResourceTypeRepositoryTemplate;
import tools.descartes.prisma.core.templates.servicebehavior.ExternalCallActionTemplate;
import tools.descartes.prisma.core.templates.servicebehavior.FineGrainedBehaviorTemplate;
import tools.descartes.prisma.core.templates.servicebehavior.ForkActionTemplate;
import tools.descartes.prisma.core.templates.servicebehavior.InternalActionTemplate;
import tools.descartes.prisma.core.templates.servicebehavior.LoopActionTemplate;
import tools.descartes.prisma.core.templates.system.SystemTemplate;

public class ModelExtractionService implements Service<ModelExtractionService> {
	
	private class CallPaths {
		private Map<String, List<String>> paths = new HashMap<>();
		
		public void addPath(String interf, String assemblyCtx) {
			List<String> a = paths.get(interf);
			if (a == null) {
				a = new LinkedList<>();
				paths.put(interf, a);
			}
			a.add(assemblyCtx);
		}
		
		public int getIndex(String interf, String assemblyCtx) {
			List<String> a = paths.get(interf);
			if (a != null) {
				int idx = 0;
				for (String cur : a) {
					if (cur.equals(assemblyCtx)) {
						return idx;
					}
					idx++;
				}
			}
			return -1;
		}
	}
	
	private static final Logger log = Logger.getLogger(ModelExtractionService.class);
	
	private InjectedValue<MonitoringService> monitoring = new InjectedValue<>();
	
	private final Map<InterfaceRecord, String> requiredRolesMapping = new HashMap<>();
	
	public static final ServiceName SERVICE_NAME = ServiceName.JBOSS.append("modelextraction", "modelextraction");
	
	public static final ModelExtractionService INSTANCE = new ModelExtractionService();

	private RepositoryConnection connection;
	
	private ApplicationModel model;
	
	private Queue<Object> updates = new ConcurrentLinkedQueue<>();
	
	private ModelUpdater updater = new ModelUpdater();
	
	private ComputingInfrastructureTemplate localHost;
	
	private String localIp;
	
	private Map<String, CallPaths> callpaths = new HashMap<>();
	
	private Thread updateThread;
	
	private ModelExtractionService() {
	}
	
	@Override
	public ModelExtractionService getValue() throws IllegalStateException, IllegalArgumentException {
		return this;
	}

	@Override
	public void start(StartContext arg0) throws StartException {
		connection = new RepositoryConnection("10.0.1.148");
		connection.open();
		model = new ApplicationModel(connection, "specjent2010");
		initModel();		
		
		if (updateThread == null) {
			updateThread = new Thread(updater);
			updateThread.start();
		}
	}

	@Override
	public void stop(StopContext arg0) {
		updater.finishUpdates();
		updateThread.interrupt();
		try {
			updateThread.join();
			updateThread = null;
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		
		try {
			connection.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	private void initModel() {
		ResourceTypeRepositoryTemplate resourceTypes = new ResourceTypeRepositoryTemplate();
		ProcessingResourceTypeTemplate cpuType = resourceTypes.defineProcessingResource("CPU");
		resourceTypes.defineProcessingResource("DELAY");
		RepositoryTemplate repository = new RepositoryTemplate();
		SystemTemplate system = new SystemTemplate();
		DistributedDataCenterTemplate ddc = new DistributedDataCenterTemplate();
		DataCenterTemplate dataCenter = new DataCenterTemplate("default");
		ddc.add(dataCenter);
		DeploymentTemplate deployment = new DeploymentTemplate();
		
		// Add a computing infrastructure corresponding to the current machine
		try {
			localIp = InetAddress.getLocalHost().getHostName();
			localHost = dataCenter.defineComputingInfrastructure(localIp);
			ActiveResourceSpecificationTemplate res = localHost.defineActiveResource("cpu-resources");
			res.defineProcessingResource("cpu", cpuType, SchedulingPolicy.PROCESSOR_SHARING, Runtime.getRuntime().availableProcessors());
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
	
		List<ModelTemplate<?,?>> changes = new LinkedList<>();
		changes.add(resourceTypes);
		changes.add(repository);
		changes.add(ddc);
		changes.add(system);
		changes.add(deployment);
		model.update(changes);		
	}
	
	public void newComponentRecord(ComponentRecord record) {
		updates.add(record);
	}
	
	public void newInterfaceRecord(InterfaceRecord record) {
		updates.add(record);
	}
	
	public void newOperationRecord(OperationRecord record) {
		updates.add(record);
	}
	
	public void newInvocationGroupRecord(InvocationGroupRecord record) {
		updates.add(record);
	}
	
	private class ModelUpdater implements Runnable {
		
		private volatile boolean stop = false;
		
		public void finishUpdates() {
			stop = true;
		}

		@Override
		public void run() {
			RepositoryTemplate repository = new RepositoryTemplate();
			SystemTemplate system = new SystemTemplate();
			DeploymentTemplate deployment = new DeploymentTemplate();

			while(!stop) {
				try {
					Thread.sleep(1000);
				} catch (InterruptedException e) {
				}
				
				Object current;
				List<ModelTemplate<?,?>> commands = new LinkedList<>();


				boolean changed = false;
				
				while((current = updates.poll()) != null) {
					if (current instanceof InvocationGroupRecord) {
						handleInvocationGroupRecord((InvocationGroupRecord)current, repository, system, deployment);
						changed = true;
					}
				}
				
				if (changed) {
					commands.add(repository);
					commands.add(system);
					commands.add(deployment);
					model.update(commands);	
				}				
			}
		}
		
		private void handleInvocationGroupRecord(InvocationGroupRecord record, RepositoryTemplate repository, SystemTemplate system, DeploymentTemplate deployment) {
			InvocationGroupRecord caller = record.getOperation().getInterface().getComponent().getCaller();			
			if (caller != null) {
				// We have information on the incoming calls as well as the outgoing calls.
				addComponent(repository, system, deployment, caller, record);
			} else {
				// no information about outgoing calls.
				addComponent(repository, system, deployment, record, null);
			}
		}		
	}
	
	private void addComponent(RepositoryTemplate repository, SystemTemplate system, DeploymentTemplate deployment, InvocationGroupRecord incomingGrec, InvocationGroupRecord outgoingGrec) {
		ComponentRecord incomingCrec = incomingGrec.getOperation().getInterface().getComponent();
		String incomingInterfaceName = incomingGrec.getOperation().getInterface().getInterfaceName();
		ComponentRecord outgoingCrec = null;
		String outgoingInterfaceName = null;
		if (outgoingGrec != null) {
			outgoingCrec = outgoingGrec.getOperation().getInterface().getComponent();		
			outgoingInterfaceName = outgoingGrec.getOperation().getInterface().getInterfaceName();
		}
		
		// 1. add the incoming interface to the repository		
		InterfaceTemplate incomingInterface = repository.defineInterface(incomingInterfaceName);
		SignatureTemplate incomingSignature = incomingInterface.defineSignature(getSignatureName(incomingGrec));
		
		// 2. add the outgoing interface to the repository (if applicable)
		InterfaceTemplate outgoingInterface = null;
		SignatureTemplate outgoingSignature = null;
		if (outgoingGrec != null) {
			outgoingInterface = repository.defineInterface(outgoingInterfaceName);
			outgoingSignature = outgoingInterface.defineSignature(getSignatureName(outgoingGrec));
		}	
		
		// 3. Create the implementation component for the caller component
		BasicComponentTemplate implementationComponent = repository.defineBasicComponent(incomingCrec.getComponentClass());
		String providingPortName = getInterfaceRoleName(incomingCrec.getComponentClass(), incomingInterface.getName(), false);
		InterfaceProvidingRoleTemplate callerProvidingRole = implementationComponent.defineInterfaceProvidingRole(providingPortName, incomingInterface);
		InterfaceRequiringRoleTemplate callerRequiringRole = null;
		int callerRoleIdx = 0;
		if (outgoingInterface != null) {
			// Here we check whether there are external calls for the same interface
			// but to different target assembly contexts. If yes, we need to distinguish the two requiring ports
			callerRoleIdx = getRequiringPortIndex(incomingCrec, outgoingGrec.getOperation().getInterface());
			String roleName = getInterfaceRoleName(incomingCrec.getComponentClass(), outgoingInterface.getName(), true);
			if (callerRoleIdx > 0) {
				roleName = roleName + ":" + callerRoleIdx;
			}
			callerRequiringRole =implementationComponent.defineInterfaceRequiringRole(roleName, outgoingInterface);
		}
		FineGrainedBehaviorTemplate behavior = createComponentInternalBehavior(implementationComponent, callerProvidingRole, incomingSignature);
		if (callerRequiringRole != null) {
			addExternalCall(incomingCrec.getComponentClass(), implementationComponent, behavior, callerRequiringRole, outgoingSignature);
		}
		RepositoryComponentTemplate<?> callerComponent = implementationComponent;
		
		// 4. Now wrap the implementation in a composite component (if necessary)
		CompositeComponentTemplate compositeComponent = createCompositeComponent(repository, incomingCrec, callerComponent);
		if (compositeComponent != null) {
			String providingRemotePortName = getInterfaceRoleName(incomingCrec.getComponentName(), incomingInterfaceName, false);			
			InterfaceProvidingRoleTemplate remoteProvidingRole = compositeComponent.defineInterfaceProvidingRole(providingRemotePortName, incomingInterface);
			InterfaceRequiringRoleTemplate remoteRequiringRole = null;
			if (outgoingInterface != null) {
				String requiringRemotePortName = getInterfaceRoleName(incomingCrec.getComponentName(), outgoingInterfaceName, false);
				if (callerRoleIdx > 0) {
					requiringRemotePortName = requiringRemotePortName + ":" + callerRoleIdx;
				}
				remoteRequiringRole = compositeComponent.defineInterfaceRequiringRole(requiringRemotePortName, outgoingInterface);			
			}
			
			wrapInCompositeComponent(incomingCrec, compositeComponent, remoteProvidingRole, remoteRequiringRole, callerComponent, callerProvidingRole, callerRequiringRole);
			callerProvidingRole = remoteProvidingRole;
			callerRequiringRole = remoteRequiringRole;
			callerComponent = compositeComponent;
		}
		
		// 5. Now add the called component (if applicable)
		InterfaceProvidingRoleTemplate calledProvidingRole = null;
		RepositoryComponentTemplate<?> calledComponent = null;
		if (outgoingGrec != null) {
			// If the component class is set on the outgoing record, then prefer this
			// (e.g., important in case of javax.jms.Queue where the component name is depending on the assembly context)
			String componentName = (outgoingCrec.getComponentClass() != null) ? outgoingCrec.getComponentClass() : outgoingCrec.getComponentName();
			calledComponent = createCalledComponent(repository, outgoingCrec);
			String calledProvingRoleName = getInterfaceRoleName(componentName, outgoingInterfaceName, false);
			calledProvidingRole = calledComponent.defineInterfaceProvidingRole(calledProvingRoleName, outgoingInterface);
		}
		
		// 6. Now create the assembly contexts
		AssemblyContextTemplate callerContext = system.defineAssemblyContext(incomingCrec.getComponentName(), callerComponent);
		if (calledComponent != null) {
			AssemblyContextTemplate calledContext = system.defineAssemblyContext(outgoingCrec.getComponentName(), calledComponent);
			system.defineAssemblyConnector(callerContext, callerRequiringRole, calledContext, calledProvidingRole);
		}
		
		// 7. Expose the provided remote interfaces as system interfaces (if this port is marked as system interface)
		String systemProvidingRoleName = localIp + "/" + incomingInterfaceName;
		// Do not create a new providing role for the system, only use the existing providing role.
		// This way a user can control which interfaces are exposed at the system level by creating them manually.
		InterfaceProvidingRoleTemplate systemProvidingRole = new InterfaceProvidingRoleTemplate(ModelTemplate.Status.EXISTS, systemProvidingRoleName, incomingInterface);
		system.defineProvidingDelegationConnector(systemProvidingRole, callerProvidingRole, callerContext);
	
		// 8. Create the deployment contexts (only for the caller component)
		deployment.defineDeploymentContext(callerContext, localHost);
	}
	
	private FineGrainedBehaviorTemplate createComponentInternalBehavior(BasicComponentTemplate implementationComponent, InterfaceProvidingRoleTemplate implementationProvidingRole, SignatureTemplate signature) {
		FineGrainedBehaviorTemplate behavior = implementationComponent.defineFineGrainedBehavior(implementationProvidingRole, signature);
		if (!isJMSComponent(implementationComponent.getName())) {
			if (behavior.getComponentInternalBehavior().getActions().isEmpty()) {
				InternalActionTemplate internalAction = new InternalActionTemplate(new ProcessingResourceTypeTemplate("CPU"));
				behavior.getComponentInternalBehavior().add(internalAction);
			}
		}
		return behavior;
	}
	
	private void addExternalCall(String componentName, 
			BasicComponentTemplate implementationComponent, 
			FineGrainedBehaviorTemplate behavior, 
			InterfaceRequiringRoleTemplate requiringRole, 
			SignatureTemplate outgoingSignature) {		
		ExternalCallActionTemplate call = new ExternalCallActionTemplate(requiringRole, outgoingSignature);
		if (isJMSComponent(componentName)) {
			if (behavior.getComponentInternalBehavior().getActions().isEmpty()) {
				ForkActionTemplate fork = new ForkActionTemplate(false, 1);
				fork.getForkedBehavior(0).add(call);
				behavior.getComponentInternalBehavior().add(fork);
			}
		} else {
			LoopActionTemplate loop = new LoopActionTemplate();
			loop.getLoopBehavior().add(call);
			behavior.getComponentInternalBehavior().add(loop);
		}
	}
	
	private CompositeComponentTemplate createCompositeComponent(RepositoryTemplate repository, ComponentRecord crec, RepositoryComponentTemplate<?> implementationComponent) {
		String compositeName = getCompositeComponentName(crec);
		if (compositeName != null) {
			return repository.defineCompositeComponent(compositeName);
		}
		return null;
	}
	
	private void wrapInCompositeComponent(ComponentRecord crec, 
			CompositeComponentTemplate compositeComponent,
			InterfaceProvidingRoleTemplate outerProvidingRole,
			InterfaceRequiringRoleTemplate outerRequiringRole,
			RepositoryComponentTemplate<?> implementationComponent,
			InterfaceProvidingRoleTemplate innerProvidingRole,
			InterfaceRequiringRoleTemplate innerRequiringRole) {
		AssemblyContextTemplate assembly = compositeComponent.defineAssemblyContext(crec.getComponentName(), implementationComponent);
		
		ProvidingDelegationConnectorTemplate providingDelegation = compositeComponent.defineProvidingDelegationConnector(outerProvidingRole, innerProvidingRole, assembly);
		
		if (innerRequiringRole != null) {
			RequiringDelegationConnectorTemplate requiringDelegation = compositeComponent.defineRequiringDelegationConnector(outerRequiringRole, innerRequiringRole, assembly);
		}
	}
	
	private RepositoryComponentTemplate<?> createCalledComponent(RepositoryTemplate repository, ComponentRecord crec) {
		String compositeName = getCompositeComponentName(crec);
		if (compositeName != null) {
			return repository.defineCompositeComponent(compositeName);
		}
		if (crec.getComponentClass() != null) {
			return repository.defineBasicComponent(crec.getComponentClass());
		}
		throw new IllegalStateException("Could not create called component for " + crec.getComponentName());
	}

	private boolean isJMSComponent(String componentName) {
		return "javax.jms.Queue".equals(componentName) || "javax.jms.Topic".equals(componentName);
	}
	
	private String getCompositeComponentName(ComponentRecord crec) {
		String componentPath = crec.getComponentName();
		if (componentPath.startsWith("/")) {
			componentPath = componentPath.substring(1);
		}
		if (componentPath.endsWith("/")) {
			componentPath = componentPath.substring(0, componentPath.length() - 1);
		}
		int idx = componentPath.lastIndexOf("/");
		if (idx < 0) {
			return null;
		} else {
			return componentPath.substring(0, idx);
		}
	}
	
	private int getRequiringPortIndex(ComponentRecord incomingCrec, InterfaceRecord outgoingIrec) {
		CallPaths paths = callpaths.get(incomingCrec.getComponentName());
		if (paths == null) {
			paths = new CallPaths();
			callpaths.put(incomingCrec.getComponentName(), paths);
		}
		paths.addPath(outgoingIrec.getInterfaceName(), outgoingIrec.getComponent().getComponentName());
		int idx = paths.getIndex(outgoingIrec.getInterfaceName(), outgoingIrec.getComponent().getComponentName());	
		if (idx < 0) {
			throw new IllegalStateException();
		}
		return idx;
	}
	
	public void extractApplicationArchitecture(ResourceTypeRepository resourceTypes, Repository repository) {
		requiredRolesMapping.clear();
		
		ResourceType cpu = null;
		ResourceType delay = null;
		for (ResourceType type : resourceTypes.getResourceTypes()) {
			if (type.getName().equals("CPU")) {
				cpu = type;
			}
			if (type.getName().equals("DELAY")) {
				delay = type;
			}
		}
		Map<String, Interface> interfaces = new HashMap<String, Interface>();
		for (Interface interf : repository.getInterfaces()) {
			interfaces.put(interf.getName(), interf);
		}
		
		for (ComponentRecord crec : monitoring.getValue().getTraceRoot().getComponents()) {
			if (crec.isFiltered()) {
				continue;
			}			
			
			BasicComponent component = getOrCreateComponent(repository, crec);
			
			addInterfaceRequiringRoles(repository, interfaces, component, crec);	

			for (InterfaceRecord irec : crec.getInterfaces()) {
				if (irec.isFiltered()) {
					continue;
				}
				
				Interface curInterface = getOrCreateInterface(interfaces, irec, repository);
				for (OperationRecord orec : irec.getOperations()) {
					if (orec.isFiltered()) {
						continue;
					}
					for (InvocationGroupRecord grec : orec.getInvocationGroups()) {
						if (grec != null) {
							addSignatureIfAbsent(curInterface, grec);
						}
					}
				}
				
				InterfaceProvidingRole providingRole = addInterfaceProvidingRole(component, irec, curInterface);
					
			
				for (OperationRecord orec : irec.getOperations()) {
					if (orec.isFiltered()) {
						continue;
					}
					for (InvocationGroupRecord grec : orec.getInvocationGroups()) {
						if (grec == null) {
							continue;
						}
						
						FineGrainedBehavior seff;
						Signature curSignature = getSignature(curInterface, grec);
						if ("javax.jms.Queue".equals(crec.getComponentClass()) || "javax.jms.Topic".equals(crec.getComponentClass())) {
							if (component.getFineGrainedBehavior().isEmpty()) {
								seff = createJmsDestinationImplementation(component, providingRole, curSignature, grec);
								component.getFineGrainedBehavior().add(seff);
							}							
						} else {
							seff = createFineGrainedBehavior(component, providingRole, curSignature, grec, cpu, delay);
							component.getFineGrainedBehavior().add(seff);
						}						
					}
				}
			}			
			repository.getComponents().add(component);
		}
	}

	private InterfaceProvidingRole addInterfaceProvidingRole(BasicComponent component, InterfaceRecord irec,
			Interface curInterface) {
		String roleName = getInterfaceRoleName(component.getName(), irec.getInterfaceName(), false);
		for (InterfaceProvidingRole role : component.getInterfaceProvidingRoles()) {
			if (role.getName().equals(roleName)) {
				return role;
			}
		}
		
		InterfaceProvidingRole providingRole = RepositoryFactory.eINSTANCE.createInterfaceProvidingRole();
		providingRole.setInterface(curInterface);
		providingRole.setName(roleName);
		component.getInterfaceProvidingRoles().add(providingRole);
		return providingRole;
	}

	private void addInterfaceRequiringRoles(Repository repository, Map<String, Interface> interfaces,
			BasicComponent component, ComponentRecord crec) {
		Map<Interface, List<InterfaceRecord>> requiredPorts = new HashMap<>();
		for (InterfaceRecord irec : crec.getInterfaces()) {
			if (crec.isFiltered()) {
				continue;
			}		
			for (OperationRecord orec : irec.getOperations()) {
				if (orec.isFiltered()) {
					continue;
				}
				
				for (InvocationGroupRecord grec : orec.getInvocationGroups()) {
					if (grec == null) {
						continue;
					}				
					for (ComponentRecord reqCrec : grec.getRequiredComponents()) {
						if (reqCrec.isFiltered()) {
							continue;
						}
	
						for (InterfaceRecord reqIrec : reqCrec.getInterfaces()) {
							if (reqIrec.isFiltered()) {
								continue;
							}								
							Interface curReqInterface = getOrCreateInterface(interfaces, reqIrec, repository);
							List<InterfaceRecord> targets = requiredPorts.get(curReqInterface);
							if (targets == null) {
								targets = new LinkedList<>();
								requiredPorts.put(curReqInterface, targets);
							}
							targets.add(reqIrec);
							
							for (OperationRecord reqOrec : reqIrec.getOperations()) {
								if (reqOrec.isFiltered()) {
									continue;
								}
								
								for (InvocationGroupRecord reqGrec : reqOrec.getInvocationGroups()) {
									if (reqGrec != null) {
										addSignatureIfAbsent(curReqInterface, reqGrec);
									}
								}
							}							
						}
					}
				}
			}
		}
		for (Interface interf : requiredPorts.keySet()) {
			Map<String, InterfaceRequiringRole> targetAssemblyContexts = new HashMap<>();
			List<InterfaceRecord> ports = requiredPorts.get(interf);
			int idx = 0;
			for (InterfaceRecord curPort : ports) {
				if (!requiredRolesMapping.containsKey(curPort)) {
					String assemblyName = curPort.getComponent().getComponentName();
					InterfaceRequiringRole requiringRole = targetAssemblyContexts.get(assemblyName);
					if (requiringRole == null) {							
						String name = getInterfaceRoleName(component.getName(), curPort.getInterfaceName(), true);
						if (idx > 0) {
							name = name + ":" + idx;
						}
						
						requiringRole = getRequiringRole(component,  name);
						if (requiringRole == null) {
							requiringRole = RepositoryFactory.eINSTANCE.createInterfaceRequiringRole();
							requiringRole.setName(name);
							requiringRole.setInterface(interf);
							component.getInterfaceRequiringRoles().add(requiringRole);
						}
						targetAssemblyContexts.put(assemblyName, requiringRole);
						idx++;
					}
					requiredRolesMapping.put(curPort, requiringRole.getName());
				}
			}
		}
	}
	
	private FineGrainedBehavior createJmsDestinationImplementation(RepositoryComponent component, InterfaceProvidingRole providingRole, Signature signature, InvocationGroupRecord grec) {
		FineGrainedBehavior seff = ServicebehaviorFactory.eINSTANCE.createFineGrainedBehavior();
		seff.setInterfaceProvidingRole(providingRole);
		seff.setDescribedSignature(signature);
		
		ComponentInternalBehavior behavior = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
		seff.setBehavior(behavior);
		
		ForkAction fork = ServicebehaviorFactory.eINSTANCE.createForkAction();
		fork.setWithSynchronizationBarrier(false);
		behavior.getActions().add(fork);
		
		ComponentInternalBehavior forkBehavior = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
		fork.getForkedBehaviors().add(forkBehavior);
		
		for (ComponentRecord reqCrec : grec.getRequiredComponents()) {
			if (reqCrec.isFiltered()) {
				continue;
			}
			for (InterfaceRecord reqIrec : reqCrec.getInterfaces()) {
				if (reqIrec.isFiltered()) {
					continue;
				}

				for (OperationRecord reqOrec : reqIrec.getOperations()) {
					if (reqOrec.isFiltered()) {
						continue;
					}
					for (InvocationGroupRecord reqGrec : reqOrec.getInvocationGroups()) {
						if (reqGrec != null && reqGrec.getInvocationCount() > 0) {
							ExternalCall extOperation = ServicebehaviorFactory.eINSTANCE.createExternalCall();
							InterfaceRequiringRole extInterfaceRole = getRequiringRole(component, requiredRolesMapping.get(reqIrec));
							extOperation.setName(getExternalCallName(reqIrec, reqOrec));
							extOperation.setInterfaceRequiringRole(extInterfaceRole);
							extOperation.setSignature(getSignature(extInterfaceRole.getInterface(), reqGrec));
							
							ExternalCallAction callAction = ServicebehaviorFactory.eINSTANCE.createExternalCallAction();
							callAction.setExternalCall(extOperation);
							forkBehavior.getActions().add(callAction);
						}
					}
				}
			}
		}
		
		return seff;
	}
	
	private FineGrainedBehavior createFineGrainedBehavior(RepositoryComponent component, InterfaceProvidingRole providingRole, Signature signature, InvocationGroupRecord grec, ResourceType cpu, ResourceType delay) {
		FineGrainedBehavior seff = ServicebehaviorFactory.eINSTANCE.createFineGrainedBehavior();
		seff.setInterfaceProvidingRole(providingRole);
		seff.setDescribedSignature(signature);
		
		ComponentInternalBehavior behavior = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
		seff.setBehavior(behavior);
		
		InternalAction internalProcessing = ServicebehaviorFactory.eINSTANCE.createInternalAction();					
		ResourceDemand cpuDemand = ServicebehaviorFactory.eINSTANCE.createResourceDemand();
		cpuDemand.setCharacterization(ModelVariableCharacterizationType.EMPIRICAL);
		cpuDemand.setResourceType(cpu);
		internalProcessing.getResourceDemand().add(cpuDemand);
		behavior.getActions().add(internalProcessing);
		
		for (ComponentRecord reqCrec : grec.getRequiredComponents()) {
			if (reqCrec.isFiltered()) {
				continue;
			}
			for (InterfaceRecord reqIrec : reqCrec.getInterfaces()) {
				if (reqIrec.isFiltered()) {
					continue;
				}

				for (OperationRecord reqOrec : reqIrec.getOperations()) {
					if (reqOrec.isFiltered()) {
						continue;
					}
					for (InvocationGroupRecord reqGrec : reqOrec.getInvocationGroups()) {
						if (reqGrec != null && reqGrec.getInvocationCount() > 0) {		
							LoopAction loop = ServicebehaviorFactory.eINSTANCE.createLoopAction();
							behavior.getActions().add(loop);
							
							LoopIterationCount invocations = ServicebehaviorFactory.eINSTANCE.createLoopIterationCount();
							invocations.setCharacterization(ModelVariableCharacterizationType.EXPLICIT);
							RandomVariable freqRV = FunctionsFactory.eINSTANCE.createRandomVariable();
							ProbabilityMassFunction pmf = FunctionsFactory.eINSTANCE.createProbabilityMassFunction();
							IntSampleList samples = FunctionsFactory.eINSTANCE.createIntSampleList();

							// Determine invocation count: currently it does not consider the distribution
							double avgCalls = ((double)reqGrec.getInvocationCount()) / grec.getInvocationCount();
							if (Math.ceil(avgCalls) == Math.floor(avgCalls)) {
								IntSample count = FunctionsFactory.eINSTANCE.createIntSample();
								count.setProbability(BigDecimal.valueOf(1.0));
								count.setValue(BigInteger.valueOf((long)avgCalls));
								samples.getItems().add(count);
							} else {									
								IntSample count = FunctionsFactory.eINSTANCE.createIntSample();
								count.setProbability(BigDecimal.valueOf(Math.ceil(avgCalls) - avgCalls));
								count.setValue(BigInteger.valueOf((long)Math.floor(avgCalls)));
								samples.getItems().add(count);
								count = FunctionsFactory.eINSTANCE.createIntSample();
								count.setProbability(BigDecimal.valueOf(1 - (Math.ceil(avgCalls) - avgCalls)));
								count.setValue(BigInteger.valueOf((long)Math.ceil(avgCalls)));
								samples.getItems().add(count);
							}
							
							pmf.setSamples(samples);
							freqRV.setProbFunction(pmf);
							invocations.setExplicitDescription(freqRV);
							loop.setLoopIterationCount(invocations);
							
							ComponentInternalBehavior loopActions = ServicebehaviorFactory.eINSTANCE.createComponentInternalBehavior();
							loop.setLoopBodyBehavior(loopActions);
							
							InternalAction networkDelay = ServicebehaviorFactory.eINSTANCE.createInternalAction();
							ResourceDemand network = ServicebehaviorFactory.eINSTANCE.createResourceDemand();
							network.setCharacterization(ModelVariableCharacterizationType.EXPLICIT);
							network.setResourceType(delay);
							RandomVariable var = FunctionsFactory.eINSTANCE.createRandomVariable();
							DoubleLiteral value = FunctionsFactory.eINSTANCE.createDoubleLiteral();
							value.setValue(BigDecimal.valueOf(0));
							var.setProbFunction(value);											
							network.setExplicitDescription(var);
							networkDelay.getResourceDemand().add(network);
							
							loopActions.getActions().add(networkDelay);
							
							ExternalCall extOperation = ServicebehaviorFactory.eINSTANCE.createExternalCall();
							InterfaceRequiringRole extInterfaceRole = getRequiringRole(component, requiredRolesMapping.get(reqIrec));
							extOperation.setName(getExternalCallName(reqIrec, reqOrec));
							extOperation.setInterfaceRequiringRole(extInterfaceRole);
							extOperation.setSignature(getSignature(extInterfaceRole.getInterface(), reqGrec));
							
							ExternalCallAction callAction = ServicebehaviorFactory.eINSTANCE.createExternalCallAction();
							callAction.setExternalCall(extOperation);
							
							loopActions.getActions().add(callAction);									
						}
					}
				}
			}
		}
		return seff;
	}
	
	public void extractSystemAssembly(Repository repository,
			edu.kit.ipd.descartes.mm.applicationlevel.system.System system) {
		for (ComponentRecord crec : monitoring.getValue().getTraceRoot().getComponents()) {
			if (crec.isFiltered()) {
				continue;
			}
			
			AssemblyContext context = getOrCreateAssemblyContext(system, crec);
			RepositoryComponent comp = context.getEncapsulatedComponent();
			if (comp == null) {
				comp = getComponent(repository, crec);				
				context.setEncapsulatedComponent(comp);
			}
			for (InterfaceRecord irec : crec.getInterfaces()) {
				if (irec.isFiltered()) {
					continue;
				}
				for (OperationRecord orec : irec.getOperations()) {
					if (orec.isFiltered()) {
						continue;
					}
					for (InvocationGroupRecord grec : orec.getInvocationGroups()) {
						if (grec == null) {
							continue;
						}
						for (ComponentRecord reqCrec : grec.getRequiredComponents()) {
							if (reqCrec.isFiltered()) {
								continue;
							}
							for (InterfaceRecord reqIrec : reqCrec.getInterfaces()) {
								if (reqIrec.isFiltered()) {
									continue;
								}
								InterfaceRequiringRole reqRole = getRequiringRole((BasicComponent)comp, requiredRolesMapping.get(reqIrec));
								AssemblyContext calledContext = getOrCreateAssemblyContext(system, reqCrec);						
								
								AssemblyConnector connector = getOrCreateAssemblyConnector(system, context, calledContext, reqRole);
								if (connector.getInterfaceProvidingRole() == null && calledContext.getEncapsulatedComponent() != null) {
									InterfaceProvidingRole role = getProvidingRole(calledContext.getEncapsulatedComponent(), reqIrec.getInterfaceName());
									connector.setInterfaceProvidingRole(role);
								}
								system.getAssemblyConnectors().add(connector);
							}
						}
					}
				}
			}
			for (AssemblyConnector connector : system.getAssemblyConnectors()) {
				if (connector.getProvidingAssemblyContext().equals(context)) {
					if (connector.getInterfaceProvidingRole() == null) {
						String interfaceName = connector.getInterfaceRequiringRole().getInterface().getName();
						connector.setInterfaceProvidingRole(getProvidingRole(comp, interfaceName));
					}
				}
			}
		}
		
	}
	
	public void extractDeployment(Repository repository,
			edu.kit.ipd.descartes.mm.applicationlevel.system.System system, Deployment deployment, String hostName) {
		DistributedDataCenter ddc = deployment.getTargetResourceLandscape();
		ComputingInfrastructure host = null;
		for (DataCenter dc : ddc.getConsistsOf()) {
			for (HardwareInfrastructure hw : dc.getContains()) {
				System.out.println(hw.getName());
				if (hw instanceof ComputingInfrastructure) {
					if (hw.getName().equals(hostName)) {
						host = (ComputingInfrastructure)hw;
						break;
					}
				}
			}
		}
		if (host == null) {
			log.error("Computing infrastructure with name: " + host +" not found.");
		}
		
		for (ComponentRecord crec : monitoring.getValue().getTraceRoot().getComponents()) {
			if (crec.isFiltered()) {
				continue;
			}
			DeploymentContext ctx = DeploymentFactory.eINSTANCE.createDeploymentContext();
			ctx.setName(crec.getComponentName() + "->" + host.getName());
			ctx.setResourceContainer(host);
			ctx.setAssemblyContext(getAssemblyContext(system, crec));
			deployment.getDeploymentContexts().add(ctx);
		}
	}
	
	public InjectedValue<MonitoringService> getMonitoringService() {
		return monitoring;
	}
	
	private AssemblyContext getAssemblyContext(edu.kit.ipd.descartes.mm.applicationlevel.system.System system, ComponentRecord record) {
		for (AssemblyContext ctx : system.getAssemblyContexts()) {
			if (ctx.getName().equals(record.getComponentName())) {
				return ctx;
			}
		}
		return null;
	}
	
	private AssemblyConnector getOrCreateAssemblyConnector(edu.kit.ipd.descartes.mm.applicationlevel.system.System system, AssemblyContext reqCtx, AssemblyContext provCtx, InterfaceRequiringRole reqRole) {
		for (AssemblyConnector con : system.getAssemblyConnectors()) {
			if (con.getRequiringAssemblyContext().equals(reqCtx)) {
				if (con.getProvidingAssemblyContext().equals(provCtx)) {
					if (con.getInterfaceRequiringRole().equals(reqRole)) {
						return con;
					}
				}
			}
		}
		
		AssemblyConnector connector = RepositoryFactory.eINSTANCE.createAssemblyConnector();
		connector.setName(reqCtx.getName() + " -> " + provCtx.getName());
		connector.setRequiringAssemblyContext(reqCtx);
		connector.setProvidingAssemblyContext(provCtx);
		connector.setInterfaceRequiringRole(reqRole);
		return connector;
	}
	
	private AssemblyContext getOrCreateAssemblyContext(edu.kit.ipd.descartes.mm.applicationlevel.system.System system, ComponentRecord record) {
		AssemblyContext existing = getAssemblyContext(system, record);
		if (existing == null) {
			AssemblyContext newAssembly = RepositoryFactory.eINSTANCE.createAssemblyContext();
			newAssembly.setId("AssemblyContext_" + getId(record.getComponentName()));
			newAssembly.setName(record.getComponentName());
			system.getAssemblyContexts().add(newAssembly);
			return newAssembly;
		}
		return existing;
	}
	
	private BasicComponent getOrCreateComponent(Repository repository, ComponentRecord crec) {
		RepositoryComponent component = getComponent(repository, crec);
		if (component == null) {
			component = RepositoryFactory.eINSTANCE.createBasicComponent();
			String name = getComponentName(crec);
			component.setId("Component_" + getId(name));
			component.setName(name);
		}
		return (BasicComponent)component;
	}
	
	private RepositoryComponent getComponent(Repository repository, ComponentRecord crec) {
		for (RepositoryComponent comp : repository.getComponents()) {
			if (comp.getName().equals(crec.getComponentClass())) {
				return comp;
			}
		}
		return null;
	}
	
	private InterfaceProvidingRole getProvidingRole(RepositoryComponent component, String interf) {
		String roleName = getInterfaceRoleName(component.getName(), interf, false);
		for (InterfaceProvidingRole role : component.getInterfaceProvidingRoles()) {
			if (role.getName().equals(roleName)) {
				return role;
			}
		}
		return null;
	}
	
	private InterfaceRequiringRole getRequiringRole(RepositoryComponent component, String name) {
		for (InterfaceRequiringRole role : component.getInterfaceRequiringRoles()) {
			if (role.getName().equals(name)) {
				return role;
			}
		}
		return null;
	}
	
	private Interface getOrCreateInterface(Map<String, Interface> existingInterfaces, InterfaceRecord crec, Repository repository) {
		Interface curInterface = existingInterfaces.get(getInterfaceName(crec));
		if (curInterface == null) {
			curInterface = RepositoryFactory.eINSTANCE.createInterface();
			String name = getInterfaceName(crec);
			curInterface.setId("Interface_" + getId(name));
			curInterface.setName(name);
			repository.getInterfaces().add(curInterface);
			existingInterfaces.put(curInterface.getName(), curInterface);
		}
		return curInterface;
	}
	
	private Signature getSignature(Interface intf, InvocationGroupRecord record) {
		String signatureName = getSignatureName(record);
		for (Signature sig : intf.getSignatures()) {
			if (sig.getName().equals(signatureName)) {
				return sig;
			}
		}
		return null;
	}
	
	private void addSignatureIfAbsent(Interface intf, InvocationGroupRecord record) {		
		Signature operationSignature = getSignature(intf, record);
		if (operationSignature == null) {
			operationSignature = RepositoryFactory.eINSTANCE.createSignature();
			String signatureId = getSignatureName(record);
			operationSignature.setId("Signature_" + getId(signatureId));
			operationSignature.setName(signatureId);
			intf.getSignatures().add(operationSignature);
		}
	}
	
	private String getSignatureName(InvocationGroupRecord record) {
		String name = record.getOperation().getInterface().getInterfaceName() + "#" + record.getOperation().getOperationName();
		if (!record.getName().isEmpty()) {
			name += "_" + record.getName();
		}
		return name;
	}
	
	private String getExternalCallName(InterfaceRecord crec, OperationRecord orec) {
		return crec.getInterfaceName() + "#" + orec.getOperationName();
	}
	
	private String getInterfaceName(InterfaceRecord crec) {
		return crec.getInterfaceName();
	}
	
	private String getInterfaceRoleName(String component, String interf, boolean req) {
		return component + "#" + (req ? "Req:" : "Prov:") + interf;
	}

	private String getComponentName(ComponentRecord crec) {
		return crec.getComponentClass();
	}
	
	private String getId(String name) {
		return name.replace("/", "_").replace(":", "_").replace(".",  "_").replace("#", "_").replace(" ", "_");
	}
}
