package tools.descartes.prisma.core.templates;

import org.eclipse.emf.cdo.eresource.CDOResource;
import org.eclipse.emf.cdo.transaction.CDOTransaction;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.Repository;
import edu.kit.ipd.descartes.mm.deployment.Deployment;
import edu.kit.ipd.descartes.mm.resourcelandscape.DataCenter;
import edu.kit.ipd.descartes.mm.resourcelandscape.DistributedDataCenter;
import edu.kit.ipd.descartes.mm.resourcetype.ResourceTypeRepository;
import edu.kit.ipd.descartes.mm.usageprofile.UsageProfile;
import edu.kit.ipd.descartes.mm.applicationlevel.system.System;
import tools.descartes.prisma.core.ApplicationModel;
import tools.descartes.prisma.core.templates.deployment.DeploymentTemplate;
import tools.descartes.prisma.core.templates.repository.RepositoryTemplate;
import tools.descartes.prisma.core.templates.resourcelandscape.DataCenterTemplate;
import tools.descartes.prisma.core.templates.resourcelandscape.DistributedDataCenterTemplate;
import tools.descartes.prisma.core.templates.resourcetype.ResourceTypeRepositoryTemplate;
import tools.descartes.prisma.core.templates.system.SystemTemplate;

public class TemplateContext {
	
	private final ApplicationModel application;
	private final CDOTransaction transaction;
	
	private final DistributedDataCenterTemplate resourcelandscape;
	private final DataCenterTemplate currentDataCenter;
	private final ResourceTypeRepositoryTemplate resourceTypes;
	private final RepositoryTemplate repository;
	private final SystemTemplate system;
	private final DeploymentTemplate deployment;
	private UsageProfile usageProfile = null;
	
	public TemplateContext(ApplicationModel application, CDOTransaction transaction) {
		this.application = application;
		this.transaction = transaction;
		
		this.resourceTypes = new ResourceTypeRepositoryTemplate();
		this.resourceTypes.setContext(this);
		
		this.repository = new RepositoryTemplate();
		this.repository.setContext(this);
		
		this.system = new SystemTemplate();
		this.system.setContext(this);
		
		this.resourcelandscape = new DistributedDataCenterTemplate();
		this.resourcelandscape.setContext(this);
		
		this.currentDataCenter = new DataCenterTemplate("default");
		
		this.deployment = new DeploymentTemplate();
		this.deployment.setContext(this);
	}
	
	public String getApplicationName() {
		return application.getApplicationName();
	}

	public CDOResource getResourceLandscapeResource() {
		return transaction.getOrCreateResource(application.getResourceLandscapePath());
	}
	
	public Repository getRepository() {
		return repository.get(null);
	}
	
	public ResourceTypeRepository getResourceTypeRepository() {
		return resourceTypes.get(null);
	}
	
	public System getSystem() {
		return system.get(null);
	}
	
	public DistributedDataCenter getDistributedDataCenter() {
		return resourcelandscape.get(null);
	}
	
	public DataCenter getCurrentDataCenter() {
		return currentDataCenter.get(getDistributedDataCenter());
	}
	
	public CDOResource getResourceTypesResource() {
		return transaction.getOrCreateResource(application.getResourceTypesPath());
	}
	
	public CDOResource getRepositoryResource() {
		return transaction.getOrCreateResource(application.getRepositoryPath());
	}
	
	public CDOResource getSystemResource() {
		return transaction.getOrCreateResource(application.getSystemPath());
	}
	
	public CDOResource getDeploymentResource() {
		return transaction.getOrCreateResource(application.getDeploymentPath());
	}
	
	public CDOResource getUsageProfileResource() {
		return transaction.getOrCreateResource(application.getUsageProfilePath());
	}

}
