package tools.descartes.prisma.core.templates.deployment;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import edu.kit.ipd.descartes.mm.applicationlevel.system.System;
import edu.kit.ipd.descartes.mm.deployment.Deployment;
import edu.kit.ipd.descartes.mm.deployment.DeploymentFactory;
import edu.kit.ipd.descartes.mm.resourcelandscape.DistributedDataCenter;
import tools.descartes.prisma.core.templates.ModelTemplate;
import tools.descartes.prisma.core.templates.repository.AssemblyContextTemplate;
import tools.descartes.prisma.core.templates.resourcelandscape.ComputingInfrastructureTemplate;

public class DeploymentTemplate extends ModelTemplate<Deployment, Deployment> {
	
	private final Set<DeploymentContextTemplate> deployments = new HashSet<>();

	public DeploymentTemplate() {
		super(null, Deployment.class);
	}
	
	public DeploymentContextTemplate defineDeploymentContext(AssemblyContextTemplate assemblyContextTemplate, ComputingInfrastructureTemplate containerTemplate) {
		DeploymentContextTemplate template = new DeploymentContextTemplate(assemblyContextTemplate, containerTemplate);
		deployments.add(template);
		return template;
	}
	
	@Override
	public Deployment get(Deployment parent) {
		Deployment ddc = find(getContext().getDeploymentResource().getContents());
		if (ddc == null) {
			ddc = create(null);
			getContext().getDeploymentResource().getContents().add(ddc);
		}
		return ddc;
	}
	

	@Override
	public List<? extends ModelTemplate<? super Deployment, ?>> getChildCommands() {
		return new ArrayList<>(deployments);
	}


	@Override
	public boolean matches(Deployment object) {
		if (object != null) {
			return getContext().getApplicationName().equals(object.getName());
		}
		return false;
	}


	@Override
	protected Deployment create(Deployment parent) {
		System system = getContext().getSystem();
		DistributedDataCenter ddc = getContext().getDistributedDataCenter();
		Deployment deployment = DeploymentFactory.eINSTANCE.createDeployment();
		deployment.setName(getContext().getApplicationName());
		deployment.setSystem(system);
		deployment.setTargetResourceLandscape(ddc);
		return deployment;
	}

}
