package tools.descartes.prisma.core.templates.repository;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ComposedStructure;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceProvidingRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryPackage;
import tools.descartes.prisma.core.templates.ModelTemplate;

public class AssemblyConnectorTemplate extends ModelTemplate<ComposedStructure, AssemblyConnector> {
	
	private final AssemblyContextTemplate addSourceContext;
	private final InterfaceRequiringRoleTemplate addRequiringRole;
	private final AssemblyContextTemplate addTargetContext;
	private final InterfaceProvidingRoleTemplate addProvidingRole;
	
	public AssemblyConnectorTemplate(AssemblyContextTemplate sourceContext, InterfaceRequiringRoleTemplate requiringRole,
			AssemblyContextTemplate targetContext, InterfaceProvidingRoleTemplate providingRole) {
		super(RepositoryPackage.Literals.COMPOSED_STRUCTURE__ASSEMBLY_CONNECTORS, AssemblyConnector.class);
		this.addSourceContext = sourceContext;
		this.addRequiringRole = requiringRole;
		this.addTargetContext = targetContext;
		this.addProvidingRole = providingRole;
	}

	@Override
	public boolean matches(AssemblyConnector connector) {
		return addSourceContext.matches(connector.getRequiringAssemblyContext())
				&& addRequiringRole.matches(connector.getInterfaceRequiringRole())
				&& addTargetContext.matches(connector.getProvidingAssemblyContext())
				&& addProvidingRole.matches(connector.getInterfaceProvidingRole());
	}

	@Override
	protected AssemblyConnector create(ComposedStructure parent) {
		AssemblyContext sourceContext = addSourceContext.get(parent);
		InterfaceRequiringRole requiringRole = addRequiringRole.get(sourceContext.getEncapsulatedComponent());
		AssemblyContext targetContext = addTargetContext.get(parent);
		InterfaceProvidingRole providingRole = addProvidingRole.get(targetContext.getEncapsulatedComponent());
		
		AssemblyConnector connector = RepositoryFactory.eINSTANCE.createAssemblyConnector();
		connector.setName(sourceContext.getName() + " -> " + targetContext.getName());
		connector.setRequiringAssemblyContext(sourceContext);
		connector.setProvidingAssemblyContext(targetContext);
		connector.setInterfaceRequiringRole(requiringRole);
		connector.setInterfaceProvidingRole(providingRole);
		return connector;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((addProvidingRole == null) ? 0 : addProvidingRole.hashCode());
		result = prime * result + ((addRequiringRole == null) ? 0 : addRequiringRole.hashCode());
		result = prime * result + ((addSourceContext == null) ? 0 : addSourceContext.hashCode());
		result = prime * result + ((addTargetContext == null) ? 0 : addTargetContext.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		AssemblyConnectorTemplate other = (AssemblyConnectorTemplate) obj;
		if (addProvidingRole == null) {
			if (other.addProvidingRole != null)
				return false;
		} else if (!addProvidingRole.equals(other.addProvidingRole))
			return false;
		if (addRequiringRole == null) {
			if (other.addRequiringRole != null)
				return false;
		} else if (!addRequiringRole.equals(other.addRequiringRole))
			return false;
		if (addSourceContext == null) {
			if (other.addSourceContext != null)
				return false;
		} else if (!addSourceContext.equals(other.addSourceContext))
			return false;
		if (addTargetContext == null) {
			if (other.addTargetContext != null)
				return false;
		} else if (!addTargetContext.equals(other.addTargetContext))
			return false;
		return true;
	}
}
