package tools.descartes.prisma.core.templates.repository;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.CompositeComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import tools.descartes.prisma.core.templates.ModelTemplate;

public class CompositeComponentTemplate extends RepositoryComponentTemplate<CompositeComponent> {
	
	private final Map<String, AssemblyContextTemplate> assemblyContexts = new HashMap<>();
	private final Set<AssemblyConnectorTemplate> assemblyConnectors = new HashSet<>();
	private final Set<ProvidingDelegationConnectorTemplate> providingDelegationConnectors = new HashSet<>();
	private final Set<RequiringDelegationConnectorTemplate> requiringDelegationConnectors = new HashSet<>();

	public CompositeComponentTemplate(String componentName) {
		super(componentName, CompositeComponent.class);
	}

	@Override
	protected CompositeComponent createRepositoryComponent() {
		return RepositoryFactory.eINSTANCE.createCompositeComponent();
	}
	
	@Override
	public List<ModelTemplate<? super CompositeComponent, ?>> getChildCommands() {
		List<ModelTemplate<? super CompositeComponent, ?>> children = super.getChildCommands();
		children.addAll(assemblyContexts.values());
		children.addAll(assemblyConnectors);
		children.addAll(providingDelegationConnectors);
		children.addAll(requiringDelegationConnectors);
		return children;
	}
	
	public AssemblyContextTemplate defineAssemblyContext(String assemblyContextName, RepositoryComponentTemplate<?> encapsulatedComponent) {
		AssemblyContextTemplate template = assemblyContexts.get(assemblyContextName);
		if (template == null) {
			template = new AssemblyContextTemplate(assemblyContextName, encapsulatedComponent);
			assemblyContexts.put(assemblyContextName, template);
		}
		return template;
	}
	
	public AssemblyConnectorTemplate defineAssemblyConnector(AssemblyContextTemplate sourceContext, InterfaceRequiringRoleTemplate requiringRole, 
			AssemblyContextTemplate targetContext, InterfaceProvidingRoleTemplate providingRole) {
		AssemblyConnectorTemplate connector = new AssemblyConnectorTemplate(sourceContext, requiringRole, targetContext, providingRole);
		assemblyConnectors.add(connector);
		return connector;
	}
	
	public ProvidingDelegationConnectorTemplate defineProvidingDelegationConnector(InterfaceProvidingRoleTemplate outerRole, InterfaceProvidingRoleTemplate innerRole, 
			AssemblyContextTemplate assemblyContext) {
		ProvidingDelegationConnectorTemplate connector = new ProvidingDelegationConnectorTemplate(outerRole, innerRole, assemblyContext);
		providingDelegationConnectors.add(connector);
		return connector;
	}
	
	public RequiringDelegationConnectorTemplate defineRequiringDelegationConnector(InterfaceRequiringRoleTemplate outerRole, InterfaceRequiringRoleTemplate innerRole, 
			AssemblyContextTemplate assemblyContext) {
		RequiringDelegationConnectorTemplate connector = new RequiringDelegationConnectorTemplate(outerRole, innerRole, assemblyContext);
		requiringDelegationConnectors.add(connector);
		return connector;
	}
	

	
}
