package tools.descartes.prisma.core.templates.repository;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ComposedProvidingRequiringEntity;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceProvidingRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ProvidingDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryPackage;
import tools.descartes.prisma.core.templates.ModelTemplate;

public class ProvidingDelegationConnectorTemplate extends ModelTemplate<ComposedProvidingRequiringEntity, ProvidingDelegationConnector> {
	
	private final InterfaceProvidingRoleTemplate outerRoleCmd;
	private final InterfaceProvidingRoleTemplate innerRoleCmd;
	private final AssemblyContextTemplate assemblyContextCmd;

	public ProvidingDelegationConnectorTemplate(InterfaceProvidingRoleTemplate outerRoleCmd, InterfaceProvidingRoleTemplate innerRoleCmd, AssemblyContextTemplate assemblyContextCmd) {
		super(RepositoryPackage.Literals.COMPOSED_STRUCTURE__PROVIDING_DELEGATION_CONNECTORS, ProvidingDelegationConnector.class);
		this.outerRoleCmd = outerRoleCmd;
		this.innerRoleCmd = innerRoleCmd;
		this.assemblyContextCmd = assemblyContextCmd;
	}

	@Override
	public boolean matches(ProvidingDelegationConnector object) {
		return outerRoleCmd.matches(object.getOuterInterfaceProvidingRole())
				&& innerRoleCmd.matches(object.getInnerInterfaceProvidingRole())
				&& assemblyContextCmd.matches(object.getAssemblyContext());
	}

	@Override
	protected ProvidingDelegationConnector create(ComposedProvidingRequiringEntity parent) {
		AssemblyContext assemblyContext = assemblyContextCmd.get(parent);
		InterfaceProvidingRole outerRole = outerRoleCmd.get(parent);
		InterfaceProvidingRole innerRole = innerRoleCmd.get(assemblyContext.getEncapsulatedComponent());
		
		if (outerRole != null && innerRole != null) {
			ProvidingDelegationConnector connector = RepositoryFactory.eINSTANCE.createProvidingDelegationConnector();
			connector.setAssemblyContext(assemblyContext);
			connector.setInnerInterfaceProvidingRole(innerRole);
			connector.setOuterInterfaceProvidingRole(outerRole);
			connector.setName(outerRole.getName() + " -> " + innerRole.getName());
			return connector;
		}
		return null;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((assemblyContextCmd == null) ? 0 : assemblyContextCmd.hashCode());
		result = prime * result + ((innerRoleCmd == null) ? 0 : innerRoleCmd.hashCode());
		result = prime * result + ((outerRoleCmd == null) ? 0 : outerRoleCmd.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		ProvidingDelegationConnectorTemplate other = (ProvidingDelegationConnectorTemplate) obj;
		if (assemblyContextCmd == null) {
			if (other.assemblyContextCmd != null)
				return false;
		} else if (!assemblyContextCmd.equals(other.assemblyContextCmd))
			return false;
		if (innerRoleCmd == null) {
			if (other.innerRoleCmd != null)
				return false;
		} else if (!innerRoleCmd.equals(other.innerRoleCmd))
			return false;
		if (outerRoleCmd == null) {
			if (other.outerRoleCmd != null)
				return false;
		} else if (!outerRoleCmd.equals(other.outerRoleCmd))
			return false;
		return true;
	}

}
