package tools.descartes.prisma.core.templates.repository;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ComposedProvidingRequiringEntity;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryPackage;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RequiringDelegationConnector;
import tools.descartes.prisma.core.templates.ModelTemplate;

public class RequiringDelegationConnectorTemplate extends ModelTemplate<ComposedProvidingRequiringEntity, RequiringDelegationConnector>{

	private final InterfaceRequiringRoleTemplate outerRoleCmd;
	private final InterfaceRequiringRoleTemplate innerRoleCmd;
	private final AssemblyContextTemplate assemblyContextCmd;

	public RequiringDelegationConnectorTemplate(InterfaceRequiringRoleTemplate outerRoleCmd, InterfaceRequiringRoleTemplate innerRoleCmd, AssemblyContextTemplate assemblyContextCmd) {
		super(RepositoryPackage.Literals.COMPOSED_STRUCTURE__REQUIRING_DELEGATION_CONNECTORS, RequiringDelegationConnector.class);
		this.innerRoleCmd = innerRoleCmd;
		this.outerRoleCmd = outerRoleCmd;
		this.assemblyContextCmd = assemblyContextCmd;
	}

	@Override
	public boolean matches(RequiringDelegationConnector object) {
		return innerRoleCmd.matches(object.getInnerInterfaceRequiringRole())
				&& outerRoleCmd.matches(object.getOuterInterfaceRequiringRole())
				&& assemblyContextCmd.matches(object.getAssemblyContext());
	}

	@Override
	protected RequiringDelegationConnector create(ComposedProvidingRequiringEntity parent) {
		AssemblyContext assemblyContext = assemblyContextCmd.get(parent);
		InterfaceRequiringRole outerRole = outerRoleCmd.get(parent);
		InterfaceRequiringRole innerRole = innerRoleCmd.get(assemblyContext.getEncapsulatedComponent());
		
		if (outerRole != null && innerRole != null) {
			RequiringDelegationConnector connector = RepositoryFactory.eINSTANCE.createRequiringDelegationConnector();
			connector.setAssemblyContext(assemblyContext);
			connector.setInnerInterfaceRequiringRole(innerRole);
			connector.setOuterInterfaceRequiringRole(outerRole);
			connector.setName(innerRole.getName() + " -> " + outerRole.getName());
			return connector;
		}
		return null;
	}
	
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((assemblyContextCmd == null) ? 0 : assemblyContextCmd.hashCode());
		result = prime * result + ((innerRoleCmd == null) ? 0 : innerRoleCmd.hashCode());
		result = prime * result + ((outerRoleCmd == null) ? 0 : outerRoleCmd.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		RequiringDelegationConnectorTemplate other = (RequiringDelegationConnectorTemplate) obj;
		if (assemblyContextCmd == null) {
			if (other.assemblyContextCmd != null)
				return false;
		} else if (!assemblyContextCmd.equals(other.assemblyContextCmd))
			return false;
		if (innerRoleCmd == null) {
			if (other.innerRoleCmd != null)
				return false;
		} else if (!innerRoleCmd.equals(other.innerRoleCmd))
			return false;
		if (outerRoleCmd == null) {
			if (other.outerRoleCmd != null)
				return false;
		} else if (!outerRoleCmd.equals(other.outerRoleCmd))
			return false;
		return true;
	}

}
