package tools.descartes.prisma.core.templates.servicebehavior;

import org.eclipse.emf.ecore.EObject;

import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCall;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import tools.descartes.prisma.core.templates.repository.InterfaceRequiringRoleTemplate;
import tools.descartes.prisma.core.templates.repository.SignatureTemplate;

public class ExternalCallActionTemplate extends AbstractActionTemplate<ExternalCallAction> {
	
	private final InterfaceRequiringRoleTemplate requiringRoleCmd;
	private final SignatureTemplate signatureCmd;
	
	public ExternalCallActionTemplate(InterfaceRequiringRoleTemplate requiringRoleCmd, SignatureTemplate signatureCmd) {
		super(ExternalCallAction.class);
		this.requiringRoleCmd = requiringRoleCmd;
		this.signatureCmd = signatureCmd;
	}

	@Override
	protected ExternalCallAction create(ComponentInternalBehavior parent) {
		RepositoryComponent comp = getComponent(parent);
		ExternalCall extOperation = ServicebehaviorFactory.eINSTANCE.createExternalCall();
		InterfaceRequiringRole extInterfaceRole = requiringRoleCmd.get(comp);
		Signature signature = signatureCmd.get(extInterfaceRole.getInterface());
		extOperation.setName(signature.getName());
		extOperation.setInterfaceRequiringRole(extInterfaceRole);
		extOperation.setSignature(signature);
		
		ExternalCallAction callAction = ServicebehaviorFactory.eINSTANCE.createExternalCallAction();
		callAction.setExternalCall(extOperation);
		return callAction;
	}
	
	@Override
	public boolean matches(ExternalCallAction a) {
		return requiringRoleCmd.matches(a.getExternalCall().getInterfaceRequiringRole())
		 			&& signatureCmd.matches(a.getExternalCall().getSignature());
	}
	
	private RepositoryComponent getComponent(ComponentInternalBehavior parent) {
		EObject current = parent;
		while(current != null) {
			if (RepositoryComponent.class.isAssignableFrom(current.getClass())) {
				return (RepositoryComponent)current;
			}
			current = current.eContainer();			
		}
		throw new IllegalStateException();
	}
}
