package tools.descartes.prisma.core.templates.servicebehavior;

import java.util.Collections;
import java.util.List;

import edu.kit.ipd.descartes.mm.applicationlevel.parameterdependencies.ModelVariableCharacterizationType;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopIterationCount;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorFactory;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ServicebehaviorPackage;
import tools.descartes.prisma.core.templates.ModelTemplate;

public class LoopActionTemplate extends AbstractActionTemplate<LoopAction> {
	
	private final ComponentInternalBehaviorTemplate<LoopAction> loopBehaviorCmd = new ComponentInternalBehaviorTemplate<>(ServicebehaviorPackage.Literals.LOOP_ACTION__LOOP_BODY_BEHAVIOR);
	
	public LoopActionTemplate() {
		super(LoopAction.class);
	}
	
	@Override
	protected LoopAction create(ComponentInternalBehavior parent) {	
		LoopAction loop = ServicebehaviorFactory.eINSTANCE.createLoopAction();		
		LoopIterationCount invocations = ServicebehaviorFactory.eINSTANCE.createLoopIterationCount();
		invocations.setCharacterization(ModelVariableCharacterizationType.EMPIRICAL);
		loop.setLoopIterationCount(invocations);
		return loop;
	}
	
	@Override
	public boolean matches(LoopAction a) {
		if (a.getLoopBodyBehavior() == null) {
			return false;
		}
		// we need to analyze the loop
		for (AbstractActionTemplate<? extends AbstractAction> actionTemplates : loopBehaviorCmd.getChildCommands()) {
			if (actionTemplates.find(a.getLoopBodyBehavior()) == null) {
				return false;
			}
		}
		return true;
	}
	
	@Override
	public List<? extends ModelTemplate<LoopAction, ?>> getChildCommands() {
		return Collections.singletonList(loopBehaviorCmd);
	}
	
	public ComponentInternalBehaviorTemplate<LoopAction> getLoopBehavior() {
		return loopBehaviorCmd;
	}
}
