package tools.descartes.prisma.core.templates.system;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import edu.kit.ipd.descartes.mm.applicationlevel.system.System;
import edu.kit.ipd.descartes.mm.applicationlevel.system.SystemFactory;
import tools.descartes.prisma.core.templates.ModelTemplate;
import tools.descartes.prisma.core.templates.repository.AssemblyConnectorTemplate;
import tools.descartes.prisma.core.templates.repository.AssemblyContextTemplate;
import tools.descartes.prisma.core.templates.repository.InterfaceProvidingRoleTemplate;
import tools.descartes.prisma.core.templates.repository.InterfaceRequiringRoleTemplate;
import tools.descartes.prisma.core.templates.repository.InterfaceTemplate;
import tools.descartes.prisma.core.templates.repository.ProvidingDelegationConnectorTemplate;
import tools.descartes.prisma.core.templates.repository.RepositoryComponentTemplate;

public class SystemTemplate extends ModelTemplate<System, System> {
	
	private final Map<String, AssemblyContextTemplate> assemblyContexts = new HashMap<>();
	private final Set<AssemblyConnectorTemplate> assemblyConnectors = new HashSet<>();
	private final Map<String, InterfaceProvidingRoleTemplate> interfaceProvidingRoles = new HashMap<>();
	private final Set<ProvidingDelegationConnectorTemplate> providingDelegationConnectors = new HashSet<>();

	public SystemTemplate() {
		super(null, System.class);
	}
	
	@Override
	public System get(System parent) {
		System system = find(getContext().getSystemResource().getContents());
		if (system == null) {
			system = create(null);
			getContext().getSystemResource().getContents().add(system);
		}
		return system;
	}
	
	public AssemblyContextTemplate defineAssemblyContext(String assemblyContextName, RepositoryComponentTemplate<?> encapsulatedComponent) {
		AssemblyContextTemplate template = assemblyContexts.get(assemblyContextName);
		if (template == null) {
			template = new AssemblyContextTemplate(assemblyContextName, encapsulatedComponent);
			assemblyContexts.put(assemblyContextName, template);
		}
		return template;
	}
	
	public AssemblyConnectorTemplate defineAssemblyConnector(AssemblyContextTemplate sourceContext, InterfaceRequiringRoleTemplate requiringRole, 
			AssemblyContextTemplate targetContext, InterfaceProvidingRoleTemplate providingRole) {
		AssemblyConnectorTemplate connector = new AssemblyConnectorTemplate(sourceContext, requiringRole, targetContext, providingRole);
		assemblyConnectors.add(connector);
		return connector;
	}
	
	public ProvidingDelegationConnectorTemplate defineProvidingDelegationConnector(InterfaceProvidingRoleTemplate outerRole, InterfaceProvidingRoleTemplate innerRole, 
			AssemblyContextTemplate assemblyContext) {
		ProvidingDelegationConnectorTemplate connector = new ProvidingDelegationConnectorTemplate(outerRole, innerRole, assemblyContext);
		providingDelegationConnectors.add(connector);
		return connector;
	}
	
	public InterfaceProvidingRoleTemplate defineInterfaceProvidingRole(String portName, InterfaceTemplate interf) {
		InterfaceProvidingRoleTemplate template = interfaceProvidingRoles.get(portName);
		if (template == null) {
			template = new InterfaceProvidingRoleTemplate(portName, interf);
			interfaceProvidingRoles.put(portName, template);
		}
		return template;
	}
	
	
	@Override
	public List<ModelTemplate<? super System, ?>> getChildCommands() {
		List<ModelTemplate<? super System,?>> children = new LinkedList<>();
		children.addAll(interfaceProvidingRoles.values());
		children.addAll(assemblyContexts.values());
		children.addAll(providingDelegationConnectors);
		children.addAll(assemblyConnectors);
		return children;
	}


	@Override
	public boolean matches(System object) {
		return object.getName().equals(getContext().getApplicationName());
	}


	@Override
	protected System create(System parent) {
		System system = SystemFactory.eINSTANCE.createSystem();
		system.setName(getContext().getApplicationName());
		return system;
	}
	
	

}
