package tools.descartes.prisma.parameterization;

import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collections;

import org.apache.log4j.BasicConfigurator;
import org.eclipse.emf.cdo.eresource.CDOResource;
import org.eclipse.emf.cdo.transaction.CDOTransaction;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.equinox.app.IApplication;
import org.eclipse.equinox.app.IApplicationContext;

import edu.kit.ipd.descartes.mm.deployment.Deployment;
import tools.descartes.librede.Librede;
import tools.descartes.librede.configuration.ConfigurationFactory;
import tools.descartes.librede.configuration.EstimationAlgorithmConfiguration;
import tools.descartes.librede.configuration.EstimationApproachConfiguration;
import tools.descartes.librede.configuration.EstimationSpecification;
import tools.descartes.librede.configuration.LibredeConfiguration;
import tools.descartes.librede.configuration.ValidationSpecification;
import tools.descartes.librede.configuration.ValidatorConfiguration;
import tools.descartes.librede.registry.Registry;
import tools.descartes.librede.units.Time;
import tools.descartes.librede.units.UnitsFactory;
import tools.descartes.prisma.core.ApplicationModel;
import tools.descartes.prisma.core.RepositoryConnection;

public class Application implements IApplication {

	@Override
	public Object start(IApplicationContext context) throws Exception {
		BasicConfigurator.configure();
		
		Librede.init();
		
		RepositoryConnection repo = new RepositoryConnection("10.0.1.148");
		ApplicationModel model = new ApplicationModel(repo, "specjent2010");
		
		Path dataPath = Paths.get("D:/Dateien/VMware/SPECjExperiments/Run_2015-09-25_11-36-26");
		
		try {
			repo.open();
			
			CDOTransaction transaction = repo.startTransaction();
			
			CDOResource res = transaction.getOrCreateResource(model.getDeploymentPath());
			if (!res.getContents().isEmpty()) {
				Path outputPath = Paths.get("C:/Users/Simon/Workspaces/runtime-Librede/prisma-output");		
				Deployment deployment = (Deployment)res.getContents().get(0);
				LibredeConfiguration conf = createLibredeConfiguration(deployment, dataPath);
				
				saveLibredeConfiguration(outputPath, conf);
			}
			
			transaction.close();
		} finally {
			repo.close();
		}
		
		return null;
	}

	@Override
	public void stop() {
		// TODO Auto-generated method stub
	}
	
	private void saveLibredeConfiguration(Path outputPath, LibredeConfiguration conf) throws IOException {
		ResourceSet rset = Registry.INSTANCE.createResourceSet();
		Resource xmi = rset.createResource(URI.createFileURI(outputPath.resolve("estimation.librede").toString()));
		xmi.getContents().add(conf);
		xmi.save(Collections.EMPTY_MAP);
	}
	
	private LibredeConfiguration createLibredeConfiguration(Deployment deployment, Path dataPath) {
		LibredeConfiguration conf = ConfigurationFactory.eINSTANCE.createLibredeConfiguration();
		
		WorkloadDescriptionDerivation workload = new WorkloadDescriptionDerivation();
		workload.updateWorkloadDescription(deployment);
		conf.setWorkloadDescription(workload.getMapping().getWorkload());
		
		InputSpecificationDerivationStrategy deriveInput = new SystemInterfaceInstrumentationStrategy();
		conf.setInput(deriveInput.derive(conf.getWorkloadDescription(), workload.getMapping(), dataPath));
//		conf.setInput(ConfigurationFactory.eINSTANCE.createInputSpecification());
		
		EstimationSpecification estimation = ConfigurationFactory.eINSTANCE.createEstimationSpecification();
		estimation.setRecursive(false);
		estimation.setWindow(60);
		estimation.setStepSize(UnitsFactory.eINSTANCE.createQuantity(1.0, Time.MINUTES));
		
		
		//Tx60
		estimation.setStartTimestamp(UnitsFactory.eINSTANCE.createQuantity(1.44316482E9, Time.SECONDS));
		estimation.setEndTimestamp(UnitsFactory.eINSTANCE.createQuantity(1.44316842E9, Time.SECONDS));
		
//		estimation.setStartTimestamp(UnitsFactory.eINSTANCE.createQuantity(1.438163220E9, Time.SECONDS));
//		estimation.setEndTimestamp(UnitsFactory.eINSTANCE.createQuantity(1.438167360E9, Time.SECONDS));
		
		//Tx100
		//estimation.setStartTimestamp(UnitsFactory.eINSTANCE.createQuantity(1.429781887E9, Time.SECONDS));
		//estimation.setEndTimestamp(UnitsFactory.eINSTANCE.createQuantity(1.429785487E9, Time.SECONDS));
		
		EstimationApproachConfiguration approach = ConfigurationFactory.eINSTANCE.createEstimationApproachConfiguration();
		approach.setType("tools.descartes.librede.approach.ServiceDemandLawApproach");
		estimation.getApproaches().add(approach);
		
		EstimationAlgorithmConfiguration algorithm = ConfigurationFactory.eINSTANCE.createEstimationAlgorithmConfiguration();
		algorithm.setType("tools.descartes.librede.algorithm.SimpleApproximation");
		estimation.getAlgorithms().add(algorithm);
		conf.setEstimation(estimation);
		
		ValidationSpecification validation = ConfigurationFactory.eINSTANCE.createValidationSpecification();
		validation.setValidateEstimates(true);
		ValidatorConfiguration rVal = ConfigurationFactory.eINSTANCE.createValidatorConfiguration();
		rVal.setType("tools.descartes.librede.validation.ResponseTimeValidator");
		validation.getValidators().add(rVal);
		ValidatorConfiguration uVal = ConfigurationFactory.eINSTANCE.createValidatorConfiguration();
		uVal.setType("tools.descartes.librede.validation.UtilizationValidator");
		validation.getValidators().add(uVal);
		
		conf.setValidation(validation);
		
		conf.setOutput(ConfigurationFactory.eINSTANCE.createOutputSpecification());
		
		return conf;
	}

}
