package tools.descartes.prisma.parameterization;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import javax.security.auth.login.Configuration;

import org.apache.log4j.Logger;

import edu.kit.ipd.descartes.mm.applicationlevel.parameterdependencies.ModelVariableCharacterizationType;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.AssemblyContext;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.BasicComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ComposedStructure;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.CompositeComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceProvidingRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.InterfaceRequiringRole;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.ProvidingDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RepositoryComponent;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.RequiringDelegationConnector;
import edu.kit.ipd.descartes.mm.applicationlevel.repository.Signature;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.AbstractAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.BranchAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ComponentInternalBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ExternalCallAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.FineGrainedBehavior;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ForkAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.InternalAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.LoopAction;
import edu.kit.ipd.descartes.mm.applicationlevel.servicebehavior.ResourceDemand;
import edu.kit.ipd.descartes.mm.applicationlevel.system.System;
import edu.kit.ipd.descartes.mm.deployment.Deployment;
import edu.kit.ipd.descartes.mm.deployment.DeploymentContext;
import edu.kit.ipd.descartes.mm.resourcelandscape.Container;
import edu.kit.ipd.descartes.mm.resourcetype.ProcessingResourceType;
import tools.descartes.librede.configuration.ConfigurationFactory;
import tools.descartes.librede.configuration.ExternalCall;
import tools.descartes.librede.configuration.ModelEntity;
import tools.descartes.librede.configuration.Resource;
import tools.descartes.librede.configuration.SchedulingStrategy;
import tools.descartes.librede.configuration.Service;

public class WorkloadDescriptionDerivation {
	
	private static final Logger log = Logger.getLogger(WorkloadDescriptionDerivation.class);

	// Set of all fully analyzed services
	private final Set<Service> completeServices = new HashSet<>();
	private MappingModel mapping = new MappingModel();
	private Map<AssemblyContext, Container> deploymentMapping = new HashMap<>();
	private Resource delayResource = null;

	public MappingModel getMapping() {
		return mapping;
	}
	
	public void updateWorkloadDescription(Deployment deployment) {
		mapping.reset();
		completeServices.clear();
		deploymentMapping.clear();
		
		for (DeploymentContext depCtx : deployment.getDeploymentContexts()) {
			deploymentMapping.put(depCtx.getAssemblyContext(), depCtx.getResourceContainer());
		}

		Stack<ComposedStructure> callStack = new Stack<ComposedStructure>();
		System system = deployment.getSystem();
		callStack.push(system);
		for (AssemblyContext ctx : system.getAssemblyContexts()) {
			Container deploymentTarget = deploymentMapping.get(ctx);
			if (deploymentTarget != null) {
				visitAssemblyContext(callStack, deploymentTarget, ctx);
			} else {
				log.warn("Assembly context " + ctx.getName() + " is not deployed on any container.");
			}			
		}
		
		// Remove all empty services
		List<Service> nonemptyServices = new LinkedList<>();
		for (Service curService : completeServices) {
			if (curService.getTasks().isEmpty()) {
				List<ExternalCall> calls = new ArrayList<>(curService.getIncomingCalls());
				for (ExternalCall curCall : calls) {
					curCall.getService().getTasks().remove(curCall);
					mapping.unmapExternalCall(curCall);
				}
				mapping.unmapService(curService);
			} else {
				nonemptyServices.add(curService);
			}
		}
		
		addDelayResource();
		
		mapping.removeUnmappedEntites();
		
	}
	
	private void addDelayResource() {
		if (delayResource == null) {
			delayResource = ConfigurationFactory.eINSTANCE.createResource();
			delayResource.setName("Delay");
			delayResource.setSchedulingStrategy(SchedulingStrategy.IS);
			delayResource.setNumberOfServers(1);
			mapping.getWorkload().getResources().add(delayResource);
		}
		
		for (ModelEntity curEntity : mapping.getNewEntities()) {
			if (curEntity instanceof Service) {
				Service curService = (Service)curEntity;
				if (curService.getIncomingCalls().isEmpty()) {
					tools.descartes.librede.configuration.ResourceDemand delayDemand = ConfigurationFactory.eINSTANCE.createResourceDemand();
					delayDemand.setName("Delay");
					delayDemand.setResource(delayResource);
					curService.getTasks().add(delayDemand);
				}
			}
		}
	}
	
	private void visitAssemblyContext(Stack<ComposedStructure> callStack, Container deploymentTarget, AssemblyContext assembly) {
		RepositoryComponent component = assembly.getEncapsulatedComponent();
		for (InterfaceProvidingRole role : assembly.getEncapsulatedComponent().getInterfaceProvidingRoles()) {
			for (Signature sig : role.getInterface().getSignatures()) {
				determineTasks(callStack, deploymentTarget, assembly, component, role, sig);
			}
		}
	}

	private void determineTasks(Stack<ComposedStructure> callStack, Container deploymentTarget, AssemblyContext assembly, RepositoryComponent component, InterfaceProvidingRole role, Signature sig) {
		if (component instanceof CompositeComponent) {
			CompositeComponent composite = (CompositeComponent)component;
			ProvidingDelegationConnector con = getDelegationConnector(composite, role);
			if (con != null) {
				RepositoryComponent inner = (RepositoryComponent)con.getInnerInterfaceProvidingRole().getInterfaceProvidingEntity();
				callStack.push(composite);
				determineTasks(callStack, deploymentTarget, con.getAssemblyContext(), inner, con.getInnerInterfaceProvidingRole(), sig);
				callStack.pop();
			} else {
				log.warn("No providing delegation connector found for role " + role.getName());
			}
		} else if (component instanceof RepositoryComponent) {
			BasicComponent implementation = (BasicComponent)component;
			FineGrainedBehavior behavior = getImplementationBehavior(implementation, role, sig);
			if (behavior != null) {
				Service service = mapping.mapService(assembly, sig, deploymentTarget);
				if (!completeServices.contains(service)) {
					visitComponentInternalBehavior(service, callStack, deploymentTarget, assembly, sig, behavior.getBehavior(), "");
					completeServices.add(service);					
					
					if (service.getName().startsWith("specj-mfg/specj-mfg/WorkOrderSession#org.spec.jent.ejb.mfg.session.WorkOrderService")) {
						// WORKAROUND: for specjent
						// The WorkOrderService interface is called internally and externally
						// Introduce an additional component to mark it as a system interface
						Service proxyService = mapping.mapService("Proxy:" + service.getName(), assembly, sig, deploymentTarget);
						if (proxyService.getTasks().isEmpty()) {
							ExternalCall call = ConfigurationFactory.eINSTANCE.createExternalCall();
							call.setCalledService(service);
							call.setName(service.getName());
							proxyService.getTasks().add(call);
						}
					}
				}
			} else {
				log.warn("No fine-grained behavior found for component " + component.getName());
			}
		}
	}
	
	private FineGrainedBehavior getImplementationBehavior(BasicComponent component, InterfaceProvidingRole role, Signature sig)	{
		for (FineGrainedBehavior currentBehavior : component.getFineGrainedBehavior()) {
			if (currentBehavior.getInterfaceProvidingRole().equals(role)
					&& currentBehavior.getDescribedSignature().equals(sig)) {
				return currentBehavior;
			}
		}
		return null;
	}
	
	private ProvidingDelegationConnector getDelegationConnector(CompositeComponent composite, InterfaceProvidingRole role) {		
		for (ProvidingDelegationConnector connector : composite.getProvidingDelegationConnectors()) {
			if (connector.getOuterInterfaceProvidingRole().equals(role)) {
				return connector;
			}
		}
		return null;
	}

	private void visitComponentInternalBehavior(Service service, Stack<ComposedStructure> callStack, Container deploymentTarget, AssemblyContext assembly, Signature signature, ComponentInternalBehavior behavior, String path) {
		int i = 0;
		path = path + "/actions.";
		for (AbstractAction action : behavior.getActions()) {
			if (action instanceof ForkAction) {
				// Only consider threads for which we will wait for completion
				// Other types of threads do not influence the response time behavior of the current service.
				if (((ForkAction)action).isWithSynchronizationBarrier()) {
					forEachComponentInternalBehavior(service, callStack, deploymentTarget, assembly, signature, ((ForkAction) action).getForkedBehaviors(), path + i +"/forkedBehaviors");
				}
			} else if (action instanceof BranchAction) {
				forEachComponentInternalBehavior(service, callStack, deploymentTarget, assembly, signature, ((BranchAction) action).getBranches(), path + i + "/branches");
			} else if (action instanceof LoopAction) {
				forEachComponentInternalBehavior(service, callStack, deploymentTarget, assembly, signature, Collections.singletonList(((LoopAction) action).getLoopBodyBehavior()), path + i + "/loopBodyBehavior");
			} else if (action instanceof InternalAction) {
				visitInternalAction(service, deploymentTarget, assembly, signature, (InternalAction)action, path + i);
			} else if (action instanceof ExternalCallAction) {
				visitExternalCallAction(service, callStack, (ExternalCallAction) action);
			}
			i++;
		}
	}
	
	private void forEachComponentInternalBehavior(Service service, Stack<ComposedStructure> callStack, Container deploymentTarget, AssemblyContext assembly, Signature signature, List<ComponentInternalBehavior> behaviors, String path) {
		int j = 0;
		for (ComponentInternalBehavior b : behaviors) {
			visitComponentInternalBehavior(service, callStack, deploymentTarget, assembly, signature, b, path + path + "." + j);
			j++;
		}
	}
	
	private void visitExternalCallAction(Service service, Stack<ComposedStructure> callStack, ExternalCallAction action) {
		InterfaceRequiringRole requiringRole = action.getExternalCall().getInterfaceRequiringRole();
		AssemblyContext providingCtx = getCalledAssemblyContext(callStack, requiringRole);
		Container targetContainer = deploymentMapping.get(providingCtx);
		
		if (providingCtx != null) {	
			Service calledService = mapping.mapService(providingCtx, action.getExternalCall().getSignature(), targetContainer);
			// Recursive calls are currently not supported.			
			if (!calledService.equals(service)) {
				mapping.mapExternalCall(service, calledService, action.getExternalCall());
			}
		} else {
			log.warn("No providing assembly context found for requiring role " + requiringRole.getName());
		}
	}
	
	private AssemblyContext getCalledAssemblyContext(Stack<ComposedStructure> callStack, InterfaceRequiringRole requiringRole) {
		ComposedStructure parent = callStack.peek();
		for (AssemblyConnector connector : parent.getAssemblyConnectors()) {
			if (connector.getInterfaceRequiringRole().equals(requiringRole)) {
				return connector.getProvidingAssemblyContext();
			}
		}
		for (RequiringDelegationConnector connector : parent.getRequiringDelegationConnectors()) {
			if (connector.getInnerInterfaceRequiringRole().equals(requiringRole)) {
				// We have to go up one level in the callstack to determine the called component
				callStack.pop();
				AssemblyContext providingCtx = getCalledAssemblyContext(callStack, connector.getOuterInterfaceRequiringRole());
				// Restore call stack
				callStack.push(parent);
				return providingCtx;
			}
		}
		
		return null;
	}
	
	private void visitInternalAction(Service parent, Container deploymentTarget, AssemblyContext assembly, Signature signature, InternalAction action, String path) {
		for (ResourceDemand demand : action.getResourceDemand()) {
			if (demand.getCharacterization() == ModelVariableCharacterizationType.EMPIRICAL) {
				if (demand.getResourceType() instanceof ProcessingResourceType) {
					Resource curResource = mapping.mapResource(deploymentTarget, demand.getResourceType());
					if (curResource == null) {
						log.warn("No processing resource of type " + demand.getResourceType() + " found in container " + deploymentTarget.getName() + ".");
						continue;
					}
					Service curService = mapping.mapService(assembly, signature, deploymentTarget);
					mapping.mapResourceDemand(curResource, curService, demand);
				}
			}
		}
	}	
		
}
 